# Copyright (c) 2013-2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
vPoller Zabbix Helper module for C clients

The vPoller Zabbix Helper module for C clients provides routines for
translating a result vPoller message to a Zabbix-friendly format.

This module is to be used by vPoller C clients only.

"""

import vpoller.helpers.zabbix


class HelperAgent(object):
    """
    HelperAgent class of the Zabbix vPoller Helper for C clients

    """
    def __init__(self, msg, data):
        """
        Initializes a new HelperAgent object

        Args:
            msg  (dict): The original request message
            data (dict): The data to be processed by the helper

        """
        self.msg = msg
        self.data = data

    def run(self):
        """
        Main Helper method

        Invokes the vpoller.helpers.zabbix helper for processing
        any data to Zabbix format and NULL-terminates the data, so
        that C clients can properly get the data we send to them.

        """
        helper = vpoller.helpers.zabbix.HelperAgent(
            msg=self.msg,
            data=self.data
        )

        result = helper.run()

        return result + '\0'



# Copyright (c) 2013-2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
vPoller Zabbix Helper module

The vPoller Zabbix Helper module provides routines for
translating a result vPoller message to a Zabbix-friendly format.

"""

import json
import logging


class HelperAgent(object):
    """
    HelperAgent class of the Zabbix vPoller Helper

    """
    def __init__(self, msg, data):
        """
        Initializes a new HelperAgent object

        Args:
            msg  (dict): The original request message
            data (dict): The data to be processed by the helper

        """
        self.msg = msg
        self.data = data

        # Methods that the Helper knows about and how to process
        self.methods = {
            'about': self.zabbix_item_value,
            'event.latest': self.zabbix_item_value,
            'session.get': self.zabbix_lld_data,
            'datacenter.discover': self.zabbix_lld_data,
            'datacenter.get': self.zabbix_item_value,
            'datacenter.alarm.get': self.zabbix_lld_data,
            'datacenter.perf.metric.get': self.zabbix_item_value,
            'datacenter.perf.metric.info': self.zabbix_lld_data,
            'cluster.alarm.get': self.zabbix_lld_data,
            'cluster.discover': self.zabbix_lld_data,
            'cluster.get': self.zabbix_item_value,
            'cluster.perf.metric.get': self.zabbix_item_value,
            'cluster.perf.metric.info': self.zabbix_lld_data,
            'host.alarm.get': self.zabbix_lld_data,
            'host.discover': self.zabbix_lld_data,
            'host.get': self.zabbix_item_value,
            'host.vm.get': self.zabbix_lld_data,
            'host.datastore.get': self.zabbix_lld_data,
            'host.cluster.get': self.zabbix_item_value,
            'host.perf.metric.get': self.zabbix_item_value,
            'host.perf.metric.info': self.zabbix_lld_data,
            'vm.alarm.get': self.zabbix_lld_data,
            'vm.discover': self.zabbix_lld_data,
            'vm.get': self.zabbix_item_value,
            'vm.datastore.get': self.zabbix_lld_data,
            'vm.disk.discover': self.zabbix_vm_disk_discover,
            'vm.disk.get': self.zabbix_vm_disk_get,
            'vm.host.get': self.zabbix_item_value,
            'vm.process.get': self.zabbix_vm_process_get,
            'vm.cpu.usage.percent': self.zabbix_item_value,
            'vm.perf.metric.get': self.zabbix_item_value,
            'vm.perf.metric.info': self.zabbix_lld_data,
            'datastore.alarm.get': self.zabbix_lld_data,
            'datastore.discover': self.zabbix_lld_data,
            'datastore.get': self.zabbix_item_value,
            'datastore.host.get': self.zabbix_lld_data,
            'datastore.vm.get': self.zabbix_lld_data,
            'datastore.perf.metric.get': self.zabbix_item_value,
            'datastore.perf.metric.info': self.zabbix_lld_data,
        }

    def run(self):
        """
        Main Helper method

        Processes the data and does any translations if needed

        """
        logging.debug('[zbx-helper]: Initiating data processing')
        logging.debug(
            '[zbx-helper]: Original client task request: %s',
            self.msg
        )
        logging.debug(
            '[zbx-helper]: Received data for processing: %s',
            self.data
        )

        # Check whether the request was successful first
        if self.data['success'] != 0:
            logging.debug(
                '[zbx-helper]: Task request was not successful (exitcode: %d)',
                self.data['success']
            )
            logging.debug(
                '[zbx-helper]: No processing will be done by the helper'
            )
            return self.data['msg']

        # The original method requested by the client
        self.method = self.msg['method']

        if self.method not in self.methods:
            logging.warning(
                '[zbx-helper]: Do not know how to process %s method',
                self.method
            )
            return '[zbx-helper]: Do not know how to process %s method' % self.method

        logging.debug(
            '[zbx-helper]: Processing data using %s() method',
            self.methods[self.method].__name__
        )
        
        result = self.methods[self.method]()

        logging.debug(
            '[zbx-helper]: Returning result after data processing: %s',
            result
        )

        return json.dumps(result, ensure_ascii=False)

    def zabbix_item_value(self):
        """
        Processes a single property value

        The value we return is of the first property only,
        so that each item in Zabbix stores a single property value.

        Returns:
            The property value from the result message

        """
        property_name = self.msg['properties'][0]
        result = self.data['result'][0]

        return result.get(property_name)

    def zabbix_vm_disk_get(self):
        """
        Processes a single property value for a VM guest disk

        The value we return is of the first property only,
        so that each item in Zabbix stores a single property value.

        Returns:
            The property value from the result message

        """
        property_name = self.msg['properties'][0]
        result = self.data['result'][0]['disk']

        return result.get(property_name)

    def zabbix_vm_disk_discover(self):
        """
        Translates a VM disk discovery request to Zabbix LLD format

        This method translates a discovery request to the
        Zabbix Low-Level Discovery format.

        For more information about Zabbix LLD,
        please refer to link below:

            - https://www.zabbix.com/documentation/2.2/manual/discovery/low_level_discovery

        The result attribute names are in Zabbix Macro-like format, e.g.

            {#VSPHERE.<TYPE>.<ATTRIBUTE>}: <value>

        """
        obj_t = self.method.split('.')[0].upper()
        result = self.data['result'][0]['disk']

        data = []

        for item in result:
            props = [('{#VSPHERE.' + obj_t + '.' + k.upper() + '}', v) for k, v in item.items()]
            data.append(dict(props))

        return {'data': data}

    def zabbix_vm_process_get(self):
        """
        Returns the number of processes in a Virtual Machine

        Matching for specific processes in the Virtual Machine is
        done by passing the 'key' attribute in the original client
        task request.

        """
        processes = self.data['result']

        if 'key' in self.msg and self.msg['key']:
            result = len([p for p in processes if self.msg['key'] in p['cmdLine']])
        else:
            result = len(processes)

        return result

    def zabbix_lld_data(self):
        """
        Translates a discovery request to Zabbix LLD format

        This method translates a discovery request to the
        Zabbix Low-Level Discovery format.

        For more information about Zabbix LLD,
        please refer to link below:

            - https://www.zabbix.com/documentation/2.2/manual/discovery/low_level_discovery

        The result attribute names are in Zabbix Macro-like format, e.g.

            {#VSPHERE.<TYPE>.<ATTRIBUTE>}: <value>

        """
        # Get the vPoller method name without the last part of it,
        # which is usually the operation performed like '.get' and
        # '.discover'. From that method name construct the macro
        # which will be used in Zabbix.
        m = self.method.split('.')[:-1]
        method_name = '.'.join(m).upper()
        result = self.data['result']

        data = []
        for item in result:
            props = [('{#VSPHERE.' + method_name + '.' + k.upper() + '}', v) for k, v in item.items()]
            data.append(dict(props))

        return {'data': data}



# Copyright (c) 2013-2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
vPoller Task Decorators module

"""

from functools import wraps
from traceback import format_exc

from vpoller.log import logger
from vpoller.task.core import Task
from vpoller.task.registry import registry

__all__ = ['task']


def task(name, required=None):
    """
    A decorator for creating new tasks

    Args:
        name      (str): Name of the task
        required (list): A list of required message keys
                         that the task expects to be present

    """
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            logger.debug('Executing task %s', name)
            try:
                result = fn(*args, **kwargs)
            except Exception as e:
                tb = format_exc()
                result = {
                    'success': 1,
                    'msg': 'Task {} failed'.format(name),
                    'traceback': tb
                }
                logger.warning('Task %s failed: %s', name, tb)
            finally:
                logger.debug('Returning result from task %s: %s', name, result)
                return result
        t = Task(name=name, function=wrapper, required=required)
        registry.register(t)
        return wrapper
    return decorator



# Copyright (c) 2013-2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
vPoller Client module for the VMware vSphere Poller

"""

import json

import zmq

from vpoller.log import logger

__all__ = ['VPollerClient', 'validate_message']


class VPollerClient(object):
    """
    VPollerClient class

    Defines methods used by clients for sending out message requests

    Sends out messages to a vPoller Proxy or vPoller Worker requesting
    properties of different vSphere objects, e.g. datastores, hosts, etc.

    Returns:
        The result message back to the client

    """
    def __init__(self, endpoint, timeout=3000, retries=3):
        """
        Initializes a VPollerClient object

        Args:
            timeout  (int): Timeout after that number of milliseconds
            retries  (int): Number of retries
            endpoint (str): Endpoint we connect the client to

        """
        self.timeout = timeout
        self.retries = retries
        self.endpoint = endpoint

    def run(self, msg):
        """
        Main vPoller Client method

        Partially based on the Lazy Pirate Pattern:

        http://zguide.zeromq.org/py:all#Client-Side-Reliability-Lazy-Pirate-Pattern

        Args:
            msg (dict): The client message to send

        """
        logger.debug('Endpoint to connect to: %s', self.endpoint)
        logger.debug('Timeout of request: %s ms', self.timeout)
        logger.debug('Number of retries: %d', self.retries)
        logger.debug('Message to be sent: %s', msg)

        self.zcontext = zmq.Context()
        self.zclient = self.zcontext.socket(zmq.REQ)
        self.zclient.connect(self.endpoint)
        self.zclient.setsockopt(zmq.LINGER, 0)
        self.zpoller = zmq.Poller()
        self.zpoller.register(self.zclient, zmq.POLLIN)
        result = None

        while self.retries > 0:
            logger.debug('Sending client message...')

            # Send our message out
            self.zclient.send_json(msg)
            socks = dict(self.zpoller.poll(self.timeout))

            # Do we have a reply?
            if socks.get(self.zclient) == zmq.POLLIN:
                logger.debug('Received response on client socket')
                result = self.zclient.recv_unicode()
                logger.debug('Received message was: %s', result)
                break
            else:
                # We didn't get a reply back from the server, let's retry
                self.retries -= 1
                logger.warning(
                    'Did not receive response, retrying...'
                )

                # Socket is confused. Close and remove it.
                logger.debug('Closing sockets and re-establishing connection...')
                self.zclient.close()
                self.zpoller.unregister(self.zclient)

                # Re-establish the connection
                logger.debug(
                    'Re-establishing connection to endpoint: %s',
                    self.endpoint
                )
                self.zclient = self.zcontext.socket(zmq.REQ)
                self.zclient.connect(self.endpoint)
                self.zclient.setsockopt(zmq.LINGER, 0)
                self.zpoller.register(self.zclient, zmq.POLLIN)

        # Close the socket and terminate the context
        logger.debug('Closing sockets and exiting')
        self.zclient.close()
        self.zpoller.unregister(self.zclient)
        self.zcontext.term()

        # Did we have any result reply at all?
        if result is None:
            logger.error(
                'Did not receive response, aborting...'
            )
            r = {
                'success': 1,
                'msg': 'Did not receive response, aborting...'
            }
            return json.dumps(r, ensure_ascii=False)

        return result


def validate_message(msg, required):
    """
    Helper method for validating a client message

    Returns:
        bool: True if message has been successfully validated

    """
    if not required:
        return True

    logger.debug(
        'Validating client message, required to have: %s',
        required
    )

    # Check if we have the required message attributes
    if not all(k in msg for k in required):
        logger.debug('Required message keys are missing')
        return False

    logger.debug('Client message successfully validated')

    return True



# Copyright (c) 2013-2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
vPoller Worker module

"""

import json
import importlib
import multiprocessing

from platform import node

try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser

import zmq

from vpoller import __version__
from vpoller.log import logger
from vpoller.client import validate_message
from vpoller.exceptions import VPollerException
from vpoller.task.registry import registry
from vconnector.core import VConnector
from vconnector.core import VConnectorDatabase

__all__ = ['VPollerWorkerManager', 'VPollerWorker']


class VPollerWorkerManager(object):
    """
    Manager of vPoller Workers

    """
    def __init__(self, config_file, num_workers=0):
        """
        Initializes a new vPoller Worker Manager

        Args:
            config_file (str): Path to the vPoller configuration file
            num_workers (str): Number of vPoller Worker 
                               processes to create

        """
        self.node = node()
        self.config_file = config_file
        self.num_workers = num_workers
        self.time_to_die = multiprocessing.Event()
        self.config = {}
        self.workers = []
        self.zcontext = None
        self.zpoller = None
        self.mgmt_socket = None
        self.mgmt_methods = {
            'status': self.status,
            'shutdown': self.signal_stop,
        }
        self.config_defaults = {
            'db': '/var/lib/vconnector/vconnector.db',
            'mgmt': 'tcp://*:10000',
            'proxy': 'tcp://localhost:10123',
            'helpers': None,
            'tasks': None,
            'cache_maxsize': 0,
            'cache_enabled': False,
            'cache_ttl': 3600,
            'cache_housekeeping': 480,
        }

    def start(self):
        """
        Start the vPoller Worker Manager and processes

        """
        logger.info('Starting Worker Manager [%s release]', __version__)

        self.load_config()
        self.create_sockets()
        self.start_workers()

        logger.info('Worker Manager is ready and running')
        while not self.time_to_die.is_set():
            try:
                self.wait_for_mgmt_task()
            except KeyboardInterrupt:
                self.signal_stop()

        self.stop()

    def stop(self):
        """
        Stop the vPoller Manager and Workers

        """
        logger.info('Worker Manager is shutting down')
        self.close_sockets()
        self.stop_workers()

    def signal_stop(self):
        """
        Signal the vPoller Worker Manager that shutdown time has arrived

        """
        logger.info('Received shutdown signal')
        self.time_to_die.set()

        return {'success': 0, 'msg': 'Shutdown time has arrived'}

    def load_config(self):
        """
        Loads the vPoller Worker Manager configuration settings

        """
        logger.debug('Loading config file %s', self.config_file)

        parser = ConfigParser(self.config_defaults)
        parser.read(self.config_file)

        self.config['mgmt'] = parser.get('worker', 'mgmt')
        self.config['db'] = parser.get('worker', 'db')
        self.config['proxy'] = parser.get('worker', 'proxy')
        self.config['helpers'] = parser.get('worker', 'helpers')
        self.config['tasks'] = parser.get('worker', 'tasks')
        self.config['cache_enabled'] = parser.getboolean('cache', 'enabled')
        self.config['cache_maxsize'] = parser.getint('cache', 'maxsize')
        self.config['cache_ttl'] = parser.getint('cache', 'ttl')
        self.config['cache_housekeeping'] = parser.getint('cache', 'housekeeping')

        if self.config['helpers']:
            self.config['helpers'] = self.config['helpers'].split(',')

        if self.config['tasks']:
            self.config['tasks'] = self.config['tasks'].split(',')
        
        logger.debug(
            'Worker Manager configuration: %s',
            self.config
        )
        
    def start_workers(self):
        """
        Start the vPoller Worker processes

        """
        logger.info('Starting Worker processes')

        if self.num_workers <= 0:
            self.num_workers = multiprocessing.cpu_count()

        logger.info(
            'Concurrency: %d (processes)',
            self.num_workers
        )

        for i in range(self.num_workers):
            worker = VPollerWorker(
                db=self.config.get('db'),
                proxy=self.config.get('proxy'),
                helpers=self.config.get('helpers'),
                tasks=self.config.get('tasks'),
                cache_enabled=self.config.get('cache_enabled'),
                cache_maxsize=self.config.get('cache_maxsize'),
                cache_ttl=self.config.get('cache_ttl'),
                cache_housekeeping=self.config.get('cache_housekeeping')
            )
            worker.daemon = True
            self.workers.append(worker)
            worker.start()

    def stop_workers(self):
        """
        Stop the vPoller Worker processes

        """
        logger.info('Stopping Worker processes')

        for worker in self.workers:
            worker.signal_stop()
            worker.join(3)

    def create_sockets(self):
        """
        Creates the ZeroMQ sockets used by the vPoller Worker Manager

        """
        logger.debug('Creating Worker Manager sockets')

        self.zcontext = zmq.Context()
        self.mgmt_socket = self.zcontext.socket(zmq.REP)
        self.mgmt_socket.bind(self.config.get('mgmt'))
        self.zpoller = zmq.Poller()
        self.zpoller.register(self.mgmt_socket, zmq.POLLIN)

    def close_sockets(self):
        """
        Closes the ZeroMQ sockets used by the Manager

        """
        logger.debug('Closing Worker Manager sockets')

        self.zpoller.unregister(self.mgmt_socket)
        self.mgmt_socket.close()
        self.zcontext.term()

    def wait_for_mgmt_task(self):
        """
        Poll the management socket for management tasks

        """
        socks = dict(self.zpoller.poll())
        if socks.get(self.mgmt_socket) == zmq.POLLIN:
            try:
                msg = self.mgmt_socket.recv_json()
            except TypeError:
                logger.warning(
                    'Invalid message received on management interface',
                )
                self.mgmt_socket.send('Invalid message received')
                return

            result = self.process_mgmt_task(msg)
            self.mgmt_socket.send_json(result)

    def process_mgmt_task(self, msg):
        """
        Processes a message for the management interface

        Example client message to shutdown the vPoller Worker would be:

            {
                "method": "shutdown"
            }

        Args:
            msg (dict): The client message for processing

        """
        logger.debug('Processing management message: %s', msg)

        if 'method' not in msg:
            return {'success': 1, 'msg': 'Missing method name'}

        if msg['method'] not in self.mgmt_methods:
            return {'success': 1, 'msg': 'Unknown method name received'}

        method = msg['method']
        result = self.mgmt_methods[method]()

        return result

    def status(self):
        """
        Get status information about the vPoller Worker

        """
        logger.debug('Getting Worker status')

        result = {
            'success': 0,
            'msg': 'vPoller Worker status',
            'result': {
                'status': 'running',
                'hostname': self.node,
                'proxy': self.config.get('proxy'),
                'mgmt': self.config.get('mgmt'),
                'db': self.config.get('db'),
                'concurrency': self.num_workers,
                'helpers': self.config.get('helpers'),
                'tasks': self.config.get('tasks'),
            }
        }

        logger.debug('Returning result to client: %s', result)

        return result


class VPollerWorker(multiprocessing.Process):
    """
    VPollerWorker class

    A vPoller Worker object runs the vSphere Agents, which are
    responsible for making vSphere API requests

    Extends:
        multiprocessing.Process

    Overrides:
        run() method

    """
    def __init__(self,
                 db,
                 proxy,
                 helpers,
                 tasks,
                 cache_enabled,
                 cache_maxsize,
                 cache_ttl,
                 cache_housekeeping,
    ):
        """
        Initialize a new VPollerWorker object

        Args:
            db                (str): Path to the vConnector database file
            proxy             (str): Endpoint to which vPoller Workers connect
                                     and receive new tasks for processing
            helpers           (list): A list of helper modules to be loaded
            task              (list): A list of task modules to be loaded
            cache_enabled     (bool): If True use an expiring cache for the
                                      managed objects
            cache_maxsize      (int): Upperbound limit on the number of items
                                      that will be stored in the cache
            cache_ttl          (int): Time in seconds after which a cached
                                      object is considered as expired
            cache_housekeeping (int): Time in minutes to perform
                                      periodic housekeeping of the cache

        """
        super(VPollerWorker, self).__init__()
        self.config = {
            'db': db,
            'proxy': proxy,
            'helpers': helpers,
            'tasks': tasks,
            'cache_enabled': cache_enabled,
            'cache_maxsize': cache_maxsize,
            'cache_ttl': cache_ttl,
            'cache_housekeeping': cache_housekeeping,
        }
        self.task_modules = {}
        self.helper_modules = {}
        self.time_to_die = multiprocessing.Event()
        self.agents = {}
        self.zcontext = None
        self.zpoller = None
        self.worker_socket = None

    def run(self):
        """
        The main worker method.

        Args:
            config (str): Path to the confuguration file for vPoller Worker

        """
        logger.info('Worker process is starting')

        self.load_task_modules()
        self.load_helper_modules()
        self.create_sockets()
        self.create_agents()

        logger.info('Worker process is ready and running')
        while not self.time_to_die.is_set():
            try:
                self.wait_for_tasks()
            except KeyboardInterrupt:
                self.signal_stop()

        self.stop()

    def stop(self):
        """
        Stop vPoller Worker process

        """
        logger.info('Worker process is shutting down')
        self.close_sockets()
        self.stop_agents()

    def signal_stop(self):
        """
        Signal the vPoller Worker process that shutdown time has arrived

        """
        self.time_to_die.set()

    def load_task_modules(self):
        """
        Loads the task modules

        """
        if not self.config.get('tasks'):
            raise VPollerException('No task modules provided')

        for task in self.config.get('tasks'):
            task = task.strip()
            logger.info('Loading task module %s', task)
            try:
                module = importlib.import_module(task)
            except ImportError as e:
                logger.warning(
                    'Cannot import task module: %s',
                    e.message
                )
                continue
            self.task_modules[task] = module

        if not self.task_modules:
            raise VPollerException('No task modules loaded')

    def load_helper_modules(self):
        """
        Loads helper modules for post-processing of results

        """
        if not self.config.get('helpers'):
            return

        for helper in self.config.get('helpers'):
            helper = helper.strip()
            logger.info('Loading helper module %s', helper)
            try:
                module = importlib.import_module(helper)
            except ImportError as e:
                logger.warning(
                    'Cannot import helper module: %s',
                    e
                )
                continue

            if not hasattr(module, 'HelperAgent'):
                logger.warning(
                    'Module %s does not provide a HelperAgent interface',
                    helper
                )
                continue

            if not hasattr(module.HelperAgent, 'run'):
                logger.warning(
                    'In module %s HelperAgent class does not provide a run() method',
                    helper
                )
                continue

            self.helper_modules[helper] = module

    def run_helper(self, helper, msg, data):
        """
        Run a helper to post-process result data

        Args:
            helper (str): Name of the helper to run
            msg   (dict): The original message request
            data  (dict): The data to be processed

        """
        logger.debug(
            'Invoking helper module %s for processing of data',
            helper
        )

        module = self.helper_modules[helper]
        h = module.HelperAgent(msg=msg, data=data)

        try:
            result = h.run()
        except Exception as e:
            logger.warning('Helper module raised an exception: %s', e.message)
            return data

        return result

    def wait_for_tasks(self):
        """
        Poll the worker socket for new tasks

        """
        socks = dict(self.zpoller.poll(1000))

        # The routing envelope of the message on the worker socket is this:
        #
        # Frame 1: [ N ][...]  <- Identity of connection
        # Frame 2: [ 0 ][]     <- Empty delimiter frame
        # Frame 3: [ N ][...]  <- Data frame
        if socks.get(self.worker_socket) == zmq.POLLIN:
            # TODO: Use recv_multipart()
            _id = self.worker_socket.recv()
            _empty = self.worker_socket.recv()

            try:
                msg = self.worker_socket.recv_json()
            except Exception as e:
                logger.warning(
                    'Invalid client message received, will be ignored',
                )
                self.worker_socket.send(_id, zmq.SNDMORE)
                self.worker_socket.send(_empty, zmq.SNDMORE)
                self.worker_socket.send_json(
                    {'success': 1, 'msg': 'Invalid message received'}
                )
                return

            # Process task and return result to client
            result = self.process_client_msg(msg)
            
            # Process data using a helper before sending it to client?
            if 'helper' in msg and msg['helper'] in self.helper_modules:
                data = self.run_helper(
                    helper=msg['helper'],
                    msg=msg,
                    data=result
                )
            else:
                # No helper specified, dump data to JSON
                try:
                    data = json.dumps(result, ensure_ascii=False)
                except ValueError as e:
                    logger.warning('Cannot serialize result: %s', e)
                    r = {
                        'success': 1,
                        'msg': 'Cannot serialize result: %s' % e
                    }
                    data = json.dumps(r)

            # Send data to client
            self.worker_socket.send(_id, zmq.SNDMORE)
            self.worker_socket.send(_empty, zmq.SNDMORE)
            try:
                self.worker_socket.send_unicode(data)
            except TypeError as e:
                logger.warning('Cannot send result: %s', e)
                r = {'success': 1, 'msg': 'Cannot send result: %s' % e}
                self.worker_socket.send_unicode(json.dumps(r))

    def create_sockets(self):
        """
        Creates the ZeroMQ sockets used by the vPoller Worker

        Creates two sockets:

        """
        logger.info('Creating Worker sockets')

        self.zcontext = zmq.Context()
        self.worker_socket = self.zcontext.socket(zmq.DEALER)
        self.worker_socket.connect(self.config.get('proxy'))
        self.zpoller = zmq.Poller()
        self.zpoller.register(self.worker_socket, zmq.POLLIN)

    def close_sockets(self):
        """
        Closes the ZeroMQ sockets used by the vPoller Worker

        """
        logger.info('Closing Worker process sockets')
        
        self.zpoller.unregister(self.worker_socket)
        self.worker_socket.close()
        self.zcontext.term()

    def create_agents(self):
        """
        Prepares the vSphere Agents used by the vPoller Worker

        Raises:
            VPollerException

        """
        logger.debug('Creating vSphere Agents')

        db = VConnectorDatabase(self.config.get('db'))
        agents = db.get_agents(only_enabled=True)

        if not agents:
            logger.warning('No registered or enabled vSphere Agents found')
            raise VPollerException(
                'No registered or enabled vSphere Agents found'
            )

        for agent in agents:
            a = VConnector(
                user=agent['user'],
                pwd=agent['pwd'],
                host=agent['host'],
                cache_enabled=self.config.get('cache_enabled'),
                cache_maxsize=self.config.get('cache_maxsize'),
                cache_ttl=self.config.get('cache_ttl'),
                cache_housekeeping=self.config.get('cache_housekeeping')
            )
            self.agents[a.host] = a
            logger.info('Created vSphere Agent for %s', agent['host'])

    def stop_agents(self):
        """
        Disconnects all vPoller Agents

        """
        logger.debug('Shutting down vSphere Agents')

        for agent in self.agents:
            self.agents[agent].disconnect()

    def process_client_msg(self, msg):
        """
        Processes a client message received on the vPoller Worker socket

        The message is passed to the VSphereAgent object of the
        respective vSphere host in order to do the actual polling.

        Args:
            msg (dict): Client message for processing

        An example message for discovering the hosts could be:

            {
                "method":   "host.discover",
                "hostname": "vc01.example.org",
            }

        An example message for polling a datastore property could be:

            {
                "method":   "datastore.poll",
                "hostname": "vc01.example.org",
                "info.url": "ds:///vmfs/volumes/5190e2a7-d2b7c58e-b1e2/",
                "property": "summary.capacity"
            }

        """
        logger.debug('Processing client message: %s', msg)

        if not isinstance(msg, dict):
            return {
                'success': 1,
                'msg': 'Expected a JSON message, received {}'.format(msg.__class__)
            }

        task = registry.get(msg.get('method'))
        agent = self.agents.get(msg.get('hostname'))

        if not task:
            return {'success': 1, 'msg': 'Unknown or missing task/method name'}

        if not agent:
            return {'success': 1, 'msg': 'Unknown or missing agent name'}

        if not validate_message(msg=msg, required=task.required):
            return {'success': 1, 'msg': 'Invalid task request'}

        result = task.function(agent, msg)

        return result



# Copyright (c) 2013-2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
vPoller Registry module

"""

from vpoller.exceptions import VPollerException
from vpoller.task.core import Task

__all__ = ['TaskRegistry']


class TaskRegistry(object):
    """
    A registry for the vPoller tasks

    """
    def __init__(self):
        self._registry = {}

    def __contains__(self, item):
        return item in self._registry

    def register(self, task):
        """
        Register a new task

        Args:
            task (Task): A Task instance to be registered

        """
        if not isinstance(task, Task):
            raise VPollerException('The task should be an instance of Task class')

        self._registry[task.name] = task

    def unregister(self, name):
        """
        Removes a task from the registry

        Args:
            name (str): Name of the task to remove from registry

        """
        self._registry.pop(name)

    def get(self, name):
        """
        Get a task by name

        """
        return self._registry.get(name)


registry = TaskRegistry()



# Copyright (c) 2013-2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
vPoller Proxy module for the VMware vSphere Poller

"""

import multiprocessing

from platform import node

try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser

import zmq

from vpoller import __version__
from vpoller.log import logger


class VPollerProxyManager(object):
    """
    Manager of vPoller Proxy

    """
    def __init__(self, config_file):
        """
        Initializes a new vPoller Proxy Manager

        Args:
            config_file (str): Path to the vPoller configuration file

        """
        self.node = node()
        self.config_file = config_file
        self.config = {}
        self.config_defaults = {
            'mgmt': 'tcp://*:9999',
            'frontend': 'tcp://*:10123',
            'backend': 'tcp://*:10124',
        }
        self.time_to_die = multiprocessing.Event()
        self.zcontext = None
        self.zpoller = None
        self.mgmt_socket = None
        self.mgmt_methods = {
            'status': self.status,
            'shutdown': self.signal_stop,
        }
        self.proxy = None

    def start(self):
        """
        Start the vPoller Proxy processes

        """
        logger.info('Starting Proxy Manager [%s release]', __version__)

        self.load_config()
        self.create_sockets()
        self.start_proxy_process()

        logger.info('Proxy Manager is ready and running')
        while not self.time_to_die.is_set():
            try:
                self.wait_for_mgmt_task()
            except KeyboardInterrupt:
                self.signal_stop()

        self.stop()

    def stop(self):
        """
        Stop the vPoller Proxy processes

        """
        logger.info('Proxy Manager is shutting down')
        self.close_sockets()
        self.stop_proxy_process()

    def signal_stop(self):
        """
        Signal the Proxy Manager that shutdown time has arrived

        """
        logger.info('Received shutdown signal')
        self.time_to_die.set()
        
        return {'success': 0, 'msg': 'Shutdown time has arrived'}

    def load_config(self):
        """
        Load the vPoller Proxy Manager configuration settings

        """
        logger.debug('Loading config file %s', self.config_file)

        parser = ConfigParser(self.config_defaults)
        parser.read(self.config_file)

        self.config['mgmt'] = parser.get('proxy', 'mgmt')
        self.config['frontend'] = parser.get('proxy', 'frontend')
        self.config['backend'] = parser.get('proxy', 'backend')

        logger.debug(
            'Proxy Manager configuration: %s',
            self.config
        )

    def start_proxy_process(self):
        """
        Start the vPoller Proxy process

        """
        logger.info('Starting Proxy process')

        self.proxy = VPollerProxy(
            frontend=self.config.get('frontend'),
            backend=self.config.get('backend')
        )
        self.proxy.daemon = True
        self.proxy.start()

    def stop_proxy_process(self):
        """
        Stop the vPoller Proxy process

        """
        logger.info('Stopping Proxy process')

        self.proxy.signal_stop()
        self.proxy.join(3)

    def create_sockets(self):
        """
        Creates the ZeroMQ sockets used by the vPoller Proxy Manager

        """
        logger.info('Creating Proxy Manager sockets')
        self.zcontext = zmq.Context()
        self.mgmt_socket = self.zcontext.socket(zmq.REP)
        self.mgmt_socket.bind(self.config.get('mgmt'))
        self.zpoller = zmq.Poller()
        self.zpoller.register(self.mgmt_socket, zmq.POLLIN)

    def close_sockets(self):
        """
        Closes the ZeroMQ sockets used by the vPoller Proxy Manager

        """
        logger.info('Closing Proxy Manager sockets')
        self.zpoller.unregister(self.mgmt_socket)
        self.mgmt_socket.close()
        self.zcontext.term()

    def wait_for_mgmt_task(self):
        """
        Poll the management socket for management tasks

        """
        socks = dict(self.zpoller.poll())
        if socks.get(self.mgmt_socket) == zmq.POLLIN:
            try:
                msg = self.mgmt_socket.recv_json()
            except TypeError:
                logger.warning(
                    'Invalid management message received',
                )
                self.mgmt_socket.send('Invalid message received')
                return

            result = self.process_mgmt_task(msg)
            self.mgmt_socket.send_json(result)

    def process_mgmt_task(self, msg):
        """
        Processes a message for the management interface

        Example client message to shutdown Proxy Manager would be:

            {
                "method": "shutdown"
            }

        Args:
            msg (dict): The client message for processing

        """
        logger.debug('Processing management message: %s', msg)

        if 'method' not in msg:
            return {
                'success': 1,
                'msg': 'Missing method name'
            }

        if msg['method'] not in self.mgmt_methods:
            return {
                'success': 1,
                'msg': 'Uknown method name received'
            }

        method = msg['method']
        result = self.mgmt_methods[method]()

        return result

    def status(self):
        """
        Get status information about the vPoller Proxy

        """
        result = {
            'success': 0,
            'msg': 'vPoller Proxy status',
            'result': {
                'status': 'running',
                'hostname': self.node,
                'mgmt': self.config.get('mgmt'),
                'frontend': self.config.get('frontend'),
                'backend': self.config.get('backend'),
            }
        }

        return result


class VPollerProxy(multiprocessing.Process):
    """
    VPollerProxy class

    A ZeroMQ proxy which distributes tasks to connected workers

    Extends:
        multiprocessing.Process

    Overrides:
        run() method

    """
    def __init__(self, frontend, backend):
        """
        Initialize a new VPollerProxy process

        Args:
            frontend (str): Endpoint to which clients connect
            backend  (str): Endpoint to which workers connect

        """
        super(VPollerProxy, self).__init__()
        self.time_to_die = multiprocessing.Event()
        self.config = {
            'frontend': frontend,
            'backend': backend,
            }
        self.zcontext = None
        self.zpoller = None
        self.frontend = None
        self.backend = None

    def run(self):
        logger.info('Proxy process is starting')

        self.create_sockets()

        logger.info('Proxy process is ready and running')
        while not self.time_to_die.is_set():
            try:
                self.distribute_tasks()
            except KeyboardInterrupt:
                self.signal_stop()

        self.stop()

    def stop(self):
        """
        Stop the vPoller Proxy process

        """
        logger.info('Proxy process is shutting down')
        self.close_sockets()

    def signal_stop(self):
        """
        Signal the vPoller Proxy process to shutdown

        """
        self.time_to_die.set()

    def distribute_tasks(self):
        """
        Distributes tasks from clients to workers for processing

        """
        socks = dict(self.zpoller.poll(1000))

        # Forward task from frontend to backend socket
        if socks.get(self.frontend) == zmq.POLLIN:
            logger.debug('Message received on frontend socket')
            task = self.frontend.recv()
            more = self.frontend.getsockopt(zmq.RCVMORE)
            if more:
                self.backend.send(task, zmq.SNDMORE)
            else:
                self.backend.send(task)
            logger.debug(
                'Sending message to backend socket: %s',
                task
            )

        # Forward result from backend to frontend socket
        if socks.get(self.backend) == zmq.POLLIN:
            logger.debug('Message received on backend socket')
            result = self.backend.recv()
            more = self.backend.getsockopt(zmq.RCVMORE)
            if more:
                self.frontend.send(result, zmq.SNDMORE)
            else:
                self.frontend.send(result)
            logger.debug(
                'Sending message to frontend socket: %s',
                result
            )

    def create_sockets(self):
        """
        Creates the ZeroMQ sockets used by the vPoller Proxy process

        """
        logger.info('Creating Proxy process sockets')
        
        self.zcontext = zmq.Context()
        self.frontend = self.zcontext.socket(zmq.ROUTER)
        self.backend = self.zcontext.socket(zmq.DEALER)
        self.frontend.bind(self.config.get('frontend'))
        self.backend.bind(self.config.get('backend'))
        self.zpoller = zmq.Poller()
        self.zpoller.register(self.frontend, zmq.POLLIN)
        self.zpoller.register(self.backend, zmq.POLLIN)

    def close_sockets(self):
        """
        Closes the ZeroMQ sockets used by vPoller Proxy process

        """
        logger.info('Closing Proxy process sockets')

        self.zpoller.unregister(self.frontend)
        self.zpoller.unregister(self.backend)
        self.frontend.close()
        self.backend.close()
        self.zcontext.term()



import re
import ast

from setuptools import setup, find_packages


_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('src/vpoller/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1))
    )

setup(
    name='vpoller',
    version=version,
    description='Distributed VMware vSphere API Proxy',
    long_description=open('README.rst').read(),
    author='Marin Atanasov Nikolov',
    author_email='dnaeon@gmail.com',
    license='BSD',
    url='https://github.com/dnaeon/py-vpoller',
    download_url='https://github.com/dnaeon/py-vpoller/releases',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    scripts=[
        'src/vpoller-client',
        'src/vpoller-proxy',
        'src/vpoller-worker',
    ],
    install_requires=[
        'pyzmq >= 14.3.1',
        'docopt >= 0.6.2',
        'pyvmomi >= 5.5.0-2014.1.1',
        'vconnector >= 0.3.7',
    ]
)



# Copyright (c) 2013-2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
VMware vSphere tasks module for vPoller

For more information about the VMware vSphere SDK,
which this module is using please check the link below:

    - https://www.vmware.com/support/developer/vc-sdk/

"""

import pyVmomi

from vpoller.log import logger
from vpoller.task.decorators import task


def _discover_objects(agent, properties, obj_type):
    """
    Helper method to simplify discovery of vSphere managed objects

    This method is used by the '*.discover' vPoller Worker methods and is
    meant for collecting properties for multiple objects at once, e.g.
    during object discovery operation.

    Args:
        agent         (VConnector): Instance of VConnector
        properties          (list): Properties to be collected
        obj_type   (pyVmomi.vim.*): Type of vSphere managed object

    Returns:
        The discovered objects in JSON format

    """
    logger.info(
        '[%s] Discovering %s managed objects',
        agent.host,
        obj_type.__name__
    )

    view_ref = agent.get_container_view(obj_type=[obj_type])
    try:
        data = agent.collect_properties(
            view_ref=view_ref,
            obj_type=obj_type,
            path_set=properties
        )
    except Exception as e:
        return {'success': 1, 'msg': 'Cannot collect properties: {}'.format(e.message)}

    view_ref.DestroyView()

    result = {
        'success': 0,
        'msg': 'Successfully discovered objects',
        'result': data,
    }

    return result

def _get_object_properties(agent,
                           properties,
                           obj_type,
                           obj_property_name,
                           obj_property_value,
                           include_mors=False):
    """
    Helper method to simplify retrieving of properties

    This method is used by the '*.get' vPoller Worker methods and is
    meant for collecting properties for a single managed object.

    We first search for the object with property name and value,
    then create a list view for this object and
    finally collect it's properties.

    Args:
        agent            (VConnector): A VConnector instance
        properties             (list): List of properties to be collected
        obj_type       pyVmomi.vim.*): Type of vSphere managed object
        obj_property_name       (str): Property name used for searching for the object
        obj_property_value      (str): Property value identifying the object in question

    Returns:
        The collected properties for this managed object in JSON format

    """
    logger.info(
        '[%s] Retrieving properties for %s managed object of type %s',
        agent.host,
        obj_property_value,
        obj_type.__name__
    )

    # Find the Managed Object reference for the requested object
    try:
        obj = agent.get_object_by_property(
            property_name=obj_property_name,
            property_value=obj_property_value,
            obj_type=obj_type
        )
    except Exception as e:
        return {'success': 1, 'msg': 'Cannot collect properties: {}'.format(e.message)}

    if not obj:
        return {
            'success': 1,
            'msg': 'Cannot find object {}'.format(obj_property_value)
        }

    # Create a list view for this object and collect properties
    view_ref = agent.get_list_view(obj=[obj])

    try:
        data = agent.collect_properties(
            view_ref=view_ref,
            obj_type=obj_type,
            path_set=properties,
            include_mors=include_mors
        )
    except Exception as e:
        return {'success': 1, 'msg': 'Cannot collect properties: {}'.format(e.message)}

    view_ref.DestroyView()

    result = {
        'success': 0,
        'msg': 'Successfully retrieved object properties',
        'result': data,
    }

    return result

def _object_datastore_get(agent, obj_type, name):
    """
    Helper method used for getting the datastores available to an object

    This method searches for the managed object with 'name' and retrieves
    the 'datastore' property which contains all datastores available/used
    by the managed object, e.g. VirtualMachine, HostSystem.

    Args:
        agent       (VConnector): A VConnector instance
        obj_type (pyVmomi.vim.*): Managed object type
        name               (str): Name of the managed object, e.g. host, vm

    Returns:
        The discovered objects in JSON format

    """
    logger.debug(
        '[%s] Getting datastores for %s managed object of type %s',
        agent.host,
        name,
        obj_type.__name__
    )

    # Find the object by it's 'name' property
    # and get the datastores available/used by it
    data = _get_object_properties(
        agent=agent,
        properties=['datastore'],
        obj_type=obj_type,
        obj_property_name='name',
        obj_property_value=name
    )

    if data['success'] != 0:
        return data

    # Get the name and datastore properties from the result
    props = data['result'][0]
    obj_datastores = props['datastore']

    # Get a list view of the datastores available/used by
    # this object and collect properties
    view_ref = agent.get_list_view(obj=obj_datastores)
    result = agent.collect_properties(
        view_ref=view_ref,
        obj_type=pyVmomi.vim.Datastore,
        path_set=['name', 'info.url']
    )

    view_ref.DestroyView()

    r = {
        'success': 0,
        'msg': 'Successfully discovered objects',
        'result': result,
    }

    return r

def _object_alarm_get(agent,
                      obj_type,
                      obj_property_name,
                      obj_property_value):
    """
    Helper method for retrieving alarms for a single Managed Object

    Args:
        agent            (VConnector): A VConnector instance
        obj_type      (pyVmomi.vim.*): Type of the Managed Object
        obj_property_name       (str): Property name used for searching for the object
        obj_property_value      (str): Property value identifying the object in question

    Returns:
        The triggered alarms for the Managed Object

    """
    logger.debug(
        '[%s] Retrieving alarms for %s managed object of type %s',
        agent.host,
        obj_property_value,
        obj_type.__name__
    )

    # Get the 'triggeredAlarmState' property for the managed object
    data = _get_object_properties(
        agent=agent,
        properties=['triggeredAlarmState'],
        obj_type=obj_type,
        obj_property_name=obj_property_name,
        obj_property_value=obj_property_value
    )

    if data['success'] != 0:
        return data

    result = []
    props = data['result'][0]
    alarms = props['triggeredAlarmState']
    for alarm in alarms:
        a = {
            'key': str(alarm.key),
            'info': alarm.alarm.info.name,
            'time': str(alarm.time),
            'entity': alarm.entity.name,
            'acknowledged': alarm.acknowledged,
            'overallStatus': alarm.overallStatus,
            'acknowledgedByUser': alarm.acknowledgedByUser,
        }
        result.append(a)

    r = {
        'success': 0,
        'msg': 'Successfully retrieved alarms',
        'result': result,
    }

    return r

def _entity_perf_metric_info(agent, entity, counter_id=None):
    """
    Get info about supported performance metrics for a managed entity

    If the managed entity supports real-time statistics then
    return the real-time performance counters available for it,
    otherwise fall back to historical statistics only.

    Args:
        agent         (VConnector): A VConnector instance
        entity     (pyVmomi.vim.*): A managed entity to lookup
        counter_id           (int): If provided return only the
                                    metrics with this counter ID

    Returns:
        Information about supported performance metrics for the entity

    """
    if not isinstance(entity, pyVmomi.vim.ManagedEntity):
        return {'success': 0, 'msg': '{} is not a managed entity'.format(entity)}

    provider_summary = agent.si.content.perfManager.QueryPerfProviderSummary(
        entity=entity
    )

    logger.debug(
        '[%s] Entity %s supports real-time statistics: %s',
        agent.host,
        entity.name,
        provider_summary.currentSupported
    )
    logger.debug(
        '[%s] Entity %s supports historical statistics: %s',
        agent.host,
        entity.name,
        provider_summary.summarySupported
    )

    interval_id = provider_summary.refreshRate if provider_summary.currentSupported else None
    try:
        metric_id = agent.si.content.perfManager.QueryAvailablePerfMetric(
            entity=entity,
            intervalId=interval_id
        )
    except pyVmomi.vim.InvalidArgument as e:
        return {
            'success': 1,
            'msg': 'Cannot retrieve performance metrics for {1}: {2}'.format(entity.name, e)
        }

    if counter_id:
        data = [{k: getattr(m, k) for k in ('counterId', 'instance')} for m in metric_id if m.counterId == counter_id]
    else:
        data = [{k: getattr(m, k) for k in ('counterId', 'instance')} for m in metric_id]

    result = {
        'msg': 'Successfully retrieved performance metrics',
        'success': 0,
        'result': data
    }

    return result

def _entity_perf_metric_get(agent, entity, counter_id, max_sample=1, instance="", perf_interval_key=None):
    """
    Retrieve performance metrics from a managed object

    Args:
        agent         (VConnector): A VConnector instance
        entity     (pyVmomi.vim.*): A managed entity (performance provider)
        counter_id           (int): A counter ID to retrieve
        max_sample           (int): Max samples to be retrieved
        instance             (str): Instance name, e.g. 'vmnic0'
        perf_interval_key    (int): Key of historical performance interval to use

    Returns:
        The collected performance metrics from the managed object

    """
    logger.info(
        '[%s] Retrieving performance metric %d for %s',
        agent.host,
        counter_id,
        entity.name,
    )

    perf_counter = [c.key for c in agent.si.content.perfManager.perfCounter] # TODO: Convert perf_counter to @property
    historical_interval = agent.si.content.perfManager.historicalInterval
    provider_summary = agent.si.content.perfManager.QueryPerfProviderSummary(
        entity=entity
    )

    logger.debug(
        '[%s] Entity %s supports real-time statistics: %s',
        agent.host,
        entity.name,
        provider_summary.currentSupported
    )
    logger.debug(
        '[%s] Entity %s supports historical statistics: %s',
        agent.host,
        entity.name,
        provider_summary.summarySupported
    )

    if not provider_summary.currentSupported and not perf_interval_key:
        logger.warning(
            '[%s] No historical performance interval provided for entity %s',
            agent.host,
            entity.name
        )
        return {'success': 1, 'msg': 'No historical performance interval provided for entity {}'.format(entity.name)}

    # For real-time statistics use the refresh rate of the provider.
    # For historical statistics use one of the existing historical
    # intervals on the system.
    # For managed entities that support both real-time and historical
    # statistics in order to retrieve historical stats a valid
    # interval key should be provided.
    # By default we expect that the requested performance counters
    # are real-time only, so if you need historical statistics
    # make sure to pass a valid historical interval key.
    if perf_interval_key:
        if perf_interval_key not in [i.key for i in historical_interval]:
            logger.warning(
                '[%s] Historical interval with key %s does not exist',
                agent.host,
                perf_interval_key
            )
            return {'success': 1, 'msg': 'Historical interval with key {} does not exist'.format(perf_interval_key)}
        else:
            interval_id = [i for i in historical_interval if i.key == perf_interval_key].pop().samplingPeriod
    else:
        interval_id = provider_summary.refreshRate

    if counter_id not in perf_counter:
        return {
            'success': 1,
            'msg': 'Unknown performance counter requested'
        }

    metric_id = pyVmomi.vim.PerformanceManager.MetricId(counterId=counter_id, instance=instance)

    # TODO: Be able to specify interval with startTime and endTime as well
    # TODO: Might want to be able to retrieve multiple metrics as well
    query_spec = pyVmomi.vim.PerformanceManager.QuerySpec(
        maxSample=max_sample,
        entity=entity,
        metricId=[metric_id],
        intervalId=interval_id
    )

    data = agent.si.content.perfManager.QueryPerf(
        querySpec=[query_spec]
    )

    result = []
    for sample in data:
        sample_info, sample_value = sample.sampleInfo, sample.value
        for value in sample_value:
            for s, v in zip(sample_info, value.value):
                d = {
                    'interval': s.interval,
                    'timestamp': str(s.timestamp),
                    'counterId': value.id.counterId,
                    'instance': value.id.instance,
                    'value': v
                }
                result.append(d)

    r = {
        'msg': 'Successfully retrieved performance metrics',
        'success': 0,
        'result': result,
    }

    return r

@task(name='about')
def about(agent, msg):
    """
    Get the 'about' information for the vSphere host

    Example client message would be:

    {
        "method":   "about",
        "hostname": "vc01.example.org"
    }

    Example client message requesting additional properties:

    {
        "method":   "about",
        "hostname": "vc01.example.org"
        "properties": [
            "apiType",
            "apiVersion",
            "version"
        ]
    }

    Returns:
        The discovered objects in JSON format

    """
    logger.info("[%s] Retrieving vSphere About information", agent.host)

    # If no properties are specified just return the 'fullName' property
    if 'properties' not in msg or not msg['properties']:
        properties = ['fullName']
    else:
        properties = msg['properties']

    data = {prop: getattr(agent.si.content.about, prop, '(null)') for prop in properties}
    result = {
        'msg': 'Successfully retrieved properties',
        'success': 0,
        'result': [data],
    }

    return result

@task(name='event.latest')
def event_latest(agent, msg):
    """
    Get the latest event registered

    Example client message would be:

    {
        "method":   "event.latest",
        "hostname": "vc01.example.org",
    }

    Returns:
        The discovered objects in JSON format

    """
    logger.info('[%s] Retrieving latest registered event', agent.host)

    e = agent.si.content.eventManager.latestEvent.fullFormattedMessage

    result = {
        'msg': 'Successfully retrieved event',
        'success': 0,
        'result': [{'event': e}],
    }

    return result

@task(name='session.get')
def session_get(agent, msg):
    """
    Get the established vSphere sessions

    Example client message would be:

    {
        "method":   "session.get",
        "hostname": "vc01.example.org",
    }

    Returns:
        The established vSphere sessions in JSON format

    """
    logger.info('[%s] Retrieving established sessions', agent.host)

    try:
        sm = agent.si.content.sessionManager
        session_list = sm.sessionList
    except pyVmomi.vim.NoPermission:
        return {
            'msg': 'No permissions to view established sessions',
            'success': 1
        }

    # Session properties to be collected
    props = [
        'key',
        'userName',
        'fullName',
        'loginTime',
        'lastActiveTime',
        'ipAddress',
        'userAgent',
        'callCount'
    ]

    sessions = []
    for session in session_list:
        s = {k: str(getattr(session, k)) for k in props}
        sessions.append(s)

    result = {
        'msg': 'Successfully retrieved sessions',
        'success': 0,
        'result': sessions,
    }

    return result

@task(name='perf.metric.info')
def perf_metric_info(agent, msg):
    """
    Get all performance counters supported by the vSphere host

    Example client message would be:

    {
        "method":   "perf.metric.info",
        "hostname": "vc01.example.org",
    }

    Returns:
        The list of supported performance counters by the vSphere host

    """
    logger.info(
        '[%s] Retrieving supported performance counters',
        agent.host
    )

    counter = agent.si.content.perfManager.perfCounter
    counter_id = [c.key for c in counter]

    try:
        counter_info = agent.si.content.perfManager.QueryPerfCounter(
            counterId=counter_id
        )
    except Exception as e:
        return {'success': 1, 'msg': 'Cannot retrieve performance metrics info: {}'.format(e.message)}

    data = []
    for c in counter_info:
        d = {
            'key': c.key,
            'nameInfo': {k: getattr(c.nameInfo, k) for k in ('label', 'summary', 'key')},
            'groupInfo': {k: getattr(c.groupInfo, k) for k in ('label', 'summary', 'key')},
            'unitInfo': {k: getattr(c.unitInfo, k) for k in ('label', 'summary', 'key')},
            'rollupType': c.rollupType,
            'statsType': c.statsType,
            'level': c.level,
            'perDeviceLevel': c.perDeviceLevel,
        }
        data.append(d)

    result = {
        'success': 0,
        'msg': 'Successfully retrieved performance metrics info',
        'result': data
    }

    return result

@task(name='perf.interval.info')
def perf_interval_info(agent, msg):
    """
    Get information about existing performance historical intervals

    Example client message would be:

    {
        "method":   "perf.interval.info",
        "hostname": "vc01.example.org",
    }

    Returns:
        The existing performance historical interval on the system

    """
    logger.info(
        '[%s] Retrieving existing performance historical intervals',
        agent.host
    )

    historical_interval = agent.si.content.perfManager.historicalInterval

    data = [{k: getattr(interval, k) for k in ('enabled', 'key', 'length', 'level', 'name', 'samplingPeriod')} for interval in historical_interval]

    result = {
        'msg': 'Successfully retrieved performance historical intervals',
        'success': 0,
        'result': data
    }

    return result

@task(name='net.discover')
def net_discover(agent, msg):
    """
    Discover all pyVmomi.vim.Network managed objects

    Example client message would be:

    {
        "method":   "net.discover",
        "hostname": "vc01.example.org",
    }

    Example client message which also requests additional properties:

    {
        "method":     "net.discover",
        "hostname":   "vc01.example.org",
        "properties": [
            "name",
            "summary.accessible"
        ]
    }

    Returns:
        The discovered objects in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    r = _discover_objects(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.Network
    )

    return r

@task(name='net.get', required=['name'])
def net_get(agent, msg):
    """
    Get properties of a single pyVmomi.vim.Network managed object

    Example client message would be:

    {
        "method":     "net.get",
        "hostname":   "vc01.example.org",
        "name":       "VM Network",
        "properties": [
            "name",
            "overallStatus"
        ]
    }

    Returns:
        The managed object properties in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    return _get_object_properties(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.Network,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

@task(name='net.host.get', required=['name'])
def net_host_get(agent, msg):
    """
    Get all Host Systems using this pyVmomi.vim.Network managed object

    Example client message would be:

    {
        "method":     "net.host.get",
        "hostname":   "vc01.example.org",
        "name":       "VM Network",
    }

    Returns:
        The managed object properties in JSON format

    """
    logger.debug(
        '[%s] Getting hosts using %s vim.Network managed object',
        agent.host,
        msg['name']
    )

    # Find the Network managed object and get the 'host' property
    data = _get_object_properties(
        agent=agent,
        properties=['name', 'host'],
        obj_type=pyVmomi.vim.Network,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    props = data['result'][0]
    network_name, network_hosts = props['name'], props['host']

    # Create a list view for the HostSystem managed objects
    view_ref = agent.get_list_view(obj=network_hosts)
    result = {}
    result['name'] = network_name
    result['host'] = agent.collect_properties(
        view_ref=view_ref,
        obj_type=pyVmomi.vim.HostSystem,
        path_set=['name']
    )

    view_ref.DestroyView()

    r = {
        'success': 0,
        'msg': 'Successfully discovered objects',
        'result': result,
    }

    return r

@task(name='net.vm.get', required=['name'])
def net_vm_get(agent, msg):
    """
    Get all Virtual Machines using this pyVmomi.vim.Network managed object

    Example client message would be:

    {
        "method":     "net.vm.get",
        "hostname":   "vc01.example.org",
        "name":       "VM Network",
    }

    Returns:
        The managed object properties in JSON format

    """
    logger.debug(
        '[%s] Getting VMs using %s vim.Network managed object',
        agent.host,
        msg['name']
    )

    # Find the Network managed object and get the 'vm' property
    data = _get_object_properties(
        agent=agent,
        properties=['name', 'vm'],
        obj_type=pyVmomi.vim.Network,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    props = data['result'][0]
    network_name, network_vms = props['name'], props['vm']

    # Create a list view for the VirtualMachine managed objects
    view_ref = agent.get_list_view(obj=network_vms)
    result = {}
    result['name'] = network_name
    result['vm'] = agent.collect_properties(
        view_ref=view_ref,
        obj_type=pyVmomi.vim.VirtualMachine,
        path_set=['name']
    )

    view_ref.DestroyView()

    r = {
        'success': 0,
        'msg': 'Successfully discovered objects',
        'result': result,
    }

    return r

@task(name='datacenter.discover')
def datacenter_discover(agent, msg):
    """
    Discover all vim.Datacenter managed objects

    Example client message would be:

    {
        "method":   "datacenter.discover",
        "hostname": "vc01.example.org",
    }

    Example client message which also requests additional properties:

    {
        "method":     "datacenter.discover",
        "hostname":   "vc01.example.org",
        "properties": [
            "name",
            "overallStatus"
        ]
    }

    Returns:
        The discovered objects in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    r = _discover_objects(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.Datacenter
    )

    return r

@task(name='datacenter.perf.metric.get', required=['name', 'counter-id'])
def datacenter_perf_metric_get(agent, msg):
    """
    Get performance metrics for a vim.Datacenter managed object

    A vim.Datacenter managed entity supports historical performance
    metrics only, so a valid historical performance interval
    should be provided as part of the client message.

    Example client message would be:

    {
        "method":   "datacenter.perf.metric.get",
        "hostname": "vc01.example.org",
        "name":     "MyDatacenter",
        "counter-id": [
            256,  # VM power on count
            257,  # VM power off count
            258   # VM suspend count
        ],
        "perf-interval": 1, # Historical performance interval '1' (Past day)
    }

    Returns:
        The retrieved performance metrics

    """
    obj = agent.get_object_by_property(
        property_name='name',
        property_value=msg['name'],
        obj_type=pyVmomi.vim.Datacenter
    )

    if not obj:
        return {'success': 1, 'msg': 'Cannot find object: %s' % msg['name']}

    try:
        counter_id = int(msg.get('counter-id'))
        perf_interval_key = int(msg.get('perf-interval')) if msg.get('perf-interval') else None
    except (TypeError, ValueError):
        logger.warning('Invalid message, cannot retrieve performance metrics')
        return {
            'success': 1,
            'msg': 'Invalid message, cannot retrieve performance metrics'
        }

    return _entity_perf_metric_get(
        agent=agent,
        entity=obj,
        counter_id=counter_id,
        perf_interval_key=perf_interval_key
    )

@task(name='datacenter.perf.metric.info')
def datacenter_perf_metric_info(agent, msg):
    """
    Get performance counters available for a vim.Datacenter object

    Example client message would be:

    {
        "method":     "datacenter.perf.metric.info",
        "hostname":   "vc01.example.org",
        "name":       "MyDatacenter",
        "counter-id": <counter-id>
    }

    Returns:
        Information about the supported performance counters for the object

    """
    obj = agent.get_object_by_property(
        property_name='name',
        property_value=msg['name'],
        obj_type=pyVmomi.vim.Datacenter
    )

    if not obj:
        return {'success': 1, 'msg': 'Cannot find object {}'.format(msg['name'])}

    counter_id = int(msg.get('counter-id')) if msg.get('counter-id') else None

    return _entity_perf_metric_info(
        agent=agent,
        entity=obj,
        counter_id=counter_id
    )

@task(name='datacenter.get', required=['name', 'properties'])
def datacenter_get(agent, msg):
    """
    Get properties of a single vim.Datacenter managed object

    Example client message would be:

    {
        "method":     "datacenter.get",
        "hostname":   "vc01.example.org",
        "name":       "MyDatacenter",
        "properties": [
            "name",
            "overallStatus"
        ]
    }

    Returns:
        The managed object properties in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    return _get_object_properties(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.Datacenter,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

@task(name='datacenter.alarm.get', required=['name'])
def datacenter_alarm_get(agent, msg):
    """
    Get all alarms for a vim.Datacenter managed object

    Example client message would be:

    {
        "method":   "datacenter.alarm.get",
        "hostname": "vc01.example.org",
        "name":     "MyDatacenter"
    }

    Returns:
        The discovered alarms in JSON format

    """
    result = _object_alarm_get(
        agent=agent,
        obj_type=pyVmomi.vim.Datacenter,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    return result

@task(name='cluster.discover')
def cluster_discover(agent, msg):
    """
    Discover all vim.ClusterComputeResource managed objects

    Example client message would be:

    {
        "method":   "cluster.discover",
        "hostname": "vc01.example.org",
    }

    Example client message which also requests additional properties:

    {
        "method":     "cluster.discover",
        "hostname":   "vc01.example.org",
        "properties": [
            "name",
            "overallStatus"
        ]
    }

    Returns:
        The discovered objects in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    r = _discover_objects(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.ClusterComputeResource
    )

    return r

@task(name='cluster.perf.metric.get', required=['name', 'counter-id'])
def cluster_perf_metric_get(agent, msg):
    """
    Get performance metrics for a vim.ClusterComputeResource managed object

    A vim.ClusterComputeResource managed entity supports historical
    statistics only, so make sure to provide a valid historical
    performance interval as part of the client message.

    Example client message would be:

    {
        "method":   "cluster.perf.metric.get",
        "hostname": "vc01.example.org",
        "name":     "MyCluster",
        "counter-id": [
            276,  # Effective memory resources
            277   # Total amount of CPU resources of all hosts in the cluster
        ],
        "perf-interval": 1, # Historical performance interval key '1' (Past day)
    }

    Returns:
        The retrieved performance metrics

    """
    obj = agent.get_object_by_property(
        property_name='name',
        property_value=msg['name'],
        obj_type=pyVmomi.vim.ClusterComputeResource
    )

    if not obj:
        return {'success': 1, 'msg': 'Cannot find object: {}'.format(msg['name'])}

    try:
        counter_id = int(msg.get('counter-id'))
        perf_interval_key = int(msg.get('perf-interval')) if msg.get('perf-interval') else None
    except (TypeError, ValueError):
        logger.warning('Invalid message, cannot retrieve performance metrics')
        return {
            'success': 1,
            'msg': 'Invalid message, cannot retrieve performance metrics'
        }

    return _entity_perf_metric_get(
        agent=agent,
        entity=obj,
        counter_id=counter_id,
        perf_interval_key=perf_interval_key
    )

@task(name='cluster.perf.metric.info')
def cluster_perf_metric_info(agent, msg):
    """
    Get performance counters available for a vim.ClusterComputeResource object

    Example client message would be:

    {
        "method":     "cluster.perf.metric.info",
        "hostname":   "vc01.example.org",
        "name":       "MyCluster",
        "counter-id": <counter-id>
    }

    Returns:
        Information about the supported performance counters for the object

    """
    obj = agent.get_object_by_property(
        property_name='name',
        property_value=msg['name'],
        obj_type=pyVmomi.vim.ClusterComputeResource
    )

    if not obj:
        return {'success': 1, 'msg': 'Cannot find object {}'.format(msg['name'])}

    counter_id = int(msg.get('counter-id')) if msg.get('counter-id') else None

    return _entity_perf_metric_info(
        agent=agent,
        entity=obj,
        counter_id=counter_id
    )

@task(name='cluster.get', required=['name', 'properties'])
def cluster_get(agent, msg):
    """
    Get properties of a vim.ClusterComputeResource managed object

    Example client message would be:

    {
        "method":     "cluster.get",
        "hostname":   "vc01.example.org",
            "name":       "MyCluster",
            "properties": [
            "name",
            "overallStatus"
        ]
    }

    Returns:
        The managed object properties in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    return _get_object_properties(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.ClusterComputeResource,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

@task(name='cluster.alarm.get', required=['name'])
def cluster_alarm_get(agent, msg):
    """
    Get all alarms for a vim.ClusterComputeResource managed object

    Example client message would be:

    {
        "method":   "cluster.alarm.get",
        "hostname": "vc01.example.org",
        "name":     "MyCluster"
    }

    Returns:
        The discovered alarms in JSON format

    """
    result = _object_alarm_get(
        agent=agent,
        obj_type=pyVmomi.vim.ClusterComputeResource,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    return result

@task(name='resource.pool.discover')
def resource_pool_discover(agent, msg):
    """
    Discover all vim.ResourcePool managed objects

    Example client message would be:

    {
        "method":   "resource.pool.discover",
        "hostname": "vc01.example.org",
    }

    Example client message which also requests additional properties:

    {
        "method":     "resource.pool.discover",
        "hostname":   "vc01.example.org",
        "properties": [
            "name",
            "overallStatus"
        ]
    }

    Returns:
        The discovered objects in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    r = _discover_objects(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.ResourcePool
    )

    return r

@task(name='resource.pool.get', required=['name', 'properties'])
def resource_pool_get(agent, msg):
    """
    Get properties of a single vim.ResourcePool managed object

    Example client message would be:

    {
        "method":     "resource.pool.get",
        "hostname":   "vc01.example.org",
        "name":       "MyResourcePool",
        "properties": [
            "name",
            "runtime.cpu",
            "runtime.memory",
            "runtime.overallStatus"
        ]
    }

    Returns:
        The managed object properties in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    return _get_object_properties(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.ResourcePool,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

@task(name='host.discover')
def host_discover(agent, msg):
    """
    Discover all vim.HostSystem managed objects

    Example client message would be:

    {
        "method":   "host.discover",
        "hostname": "vc01.example.org",
    }

    Example client message which also requests additional properties:

    {
        "method":     "host.discover",
        "hostname":   "vc01.example.org",
        "properties": [
            "name",
            "runtime.powerState"
        ]
    }

    Returns:
        The discovered objects in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    r = _discover_objects(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.HostSystem
    )

    return r

@task(name='host.get', required=['name', 'properties'])
def host_get(agent, msg):
    """
    Get properties of a single vim.HostSystem managed object

    Example client message would be:

    {
        "method":     "host.get",
        "hostname":   "vc01.example.org",
        "name":       "esxi01.example.org",
        "properties": [
            "name",
            "runtime.powerState"
        ]
    }

    Returns:
        The managed object properties in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    return _get_object_properties(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.HostSystem,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

@task(name='host.alarm.get', required=['name'])
def host_alarm_get(agent, msg):
    """
    Get all alarms for a vim.HostSystem managed object

    Example client message would be:

    {
        "method":   "host.alarm.get",
        "hostname": "vc01.example.org",
        "name":     "esxi01.example.org"
    }

    Returns:
        The discovered alarms in JSON format

    """
    result = _object_alarm_get(
        agent=agent,
        obj_type=pyVmomi.vim.HostSystem,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    return result

@task(name='host.perf.metric.get', required=['name', 'counter-id'])
def host_perf_metric_get(agent, msg):
    """
    Get performance metrics for a vim.HostSystem managed object

    Example client message would be:

    {
        "method":   "host.perf.metric.get",
        "hostname": "vc01.example.org",
        "name":     "esxi01.example.org",
        "properties": [
            97,  # Amount of host physical memory consumed by a virtual machine, host, or cluster
            130  # Average number of kilobytes read from the disk each second during the collection interval
        ],
        "max_sample": 1
    }

    Returns:
        The retrieved performance metrics

    """
    obj = agent.get_object_by_property(
        property_name='name',
        property_value=msg['name'],
        obj_type=pyVmomi.vim.HostSystem
    )

    if not obj:
        return {'success': 1, 'msg': 'Cannot find object: {}'.format(msg['name'])}

    try:
        counter_id = int(msg.get('counter-id'))
        max_sample = int(msg.get('max-sample')) if msg.get('max-sample') else 1
        perf_interval_key = int(msg.get('perf-interval')) if msg.get('perf-interval') else None
        instance = msg.get('instance') if msg.get('instance') else ''
    except (TypeError, ValueError):
        logger.warning('Invalid message, cannot retrieve performance metrics')
        return {
            'success': 1,
            'msg': 'Invalid message, cannot retrieve performance metrics'
        }

    return _entity_perf_metric_get(
        agent=agent,
        entity=obj,
        counter_id=counter_id,
        max_sample=max_sample,
        instance=instance,
        perf_interval_key=perf_interval_key
    )

@task(name='host.perf.metric.info', required=['name'])
def host_perf_metric_info(agent, msg):
    """
    Get performance counters available for a vim.HostSystem object

    Example client message would be:

    {
        "method":     "host.perf.metric.info",
        "hostname":   "vc01.example.org",
        "name":       "esxi01.example.org",
        "counter-id":  <counter-id>
    }

    Returns:
        Information about the supported performance counters for the object

    """
    obj = agent.get_object_by_property(
        property_name='name',
        property_value=msg['name'],
        obj_type=pyVmomi.vim.HostSystem
    )

    if not obj:
        return {'success': 1, 'msg': 'Cannot find object {}'.format(msg['name'])}

    counter_id = int(msg.get('counter-id')) if msg.get('counter-id') else None

    return _entity_perf_metric_info(
        agent=agent,
        entity=obj,
        counter_id=counter_id
    )

@task(name='host.cluster.get', required=['name'])
def host_cluster_get(agent, msg):
    """
    Get the cluster name for a HostSystem

    Example client message would be:

    {
        "method":     "host.cluster.get",
        "hostname":   "vc01.example.org",
        "name":       "esxi01.example.org",
    }

    Returns:
        The managed object properties in JSON format

    """
    logger.debug(
        '[%s] Getting cluster name for %s host',
        agent.host,
        msg['name']
    )

    # Find the HostSystem managed object and get the 'parent' property
    data = _get_object_properties(
        agent=agent,
        properties=['name', 'parent'],
        obj_type=pyVmomi.vim.HostSystem,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    props = data['result'][0]
    host_name, host_cluster = props['name'], props['parent']

    result = {}
    result['name'] = host_name
    result['cluster'] = host_cluster.name

    r = {
        'success': 0,
        'msg': 'Successfully retrieved properties',
        'result': [result],
    }

    return r

@task(name='host.vm.get', required=['name'])
def host_vm_get(agent, msg):
    """
    Get all vim.VirtualMachine objects of a HostSystem

    Example client message would be:

        {
            "method":     "host.vm.get",
            "hostname":   "vc01.example.org",
            "name":       "esxi01.example.org",
        }

    Returns:
        The managed object properties in JSON format

    """
    logger.debug(
        '[%s] Getting VirtualMachine list running on %s host',
        agent.host,
        msg['name']
    )

    # Find the HostSystem managed object and get the 'vm' property
    data = _get_object_properties(
        agent=agent,
        properties=['vm'],
        obj_type=pyVmomi.vim.HostSystem,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    props = data['result'][0]
    host_vms = props['vm']

    # Create a list view for the VirtualMachine managed objects
    view_ref = agent.get_list_view(obj=host_vms)
    result = agent.collect_properties(
        view_ref=view_ref,
        obj_type=pyVmomi.vim.VirtualMachine,
        path_set=['name']
    )

    view_ref.DestroyView()

    r = {
        'success': 0,
        'msg': 'Successfully discovered objects',
        'result': result,
    }

    return r

@task(name='host.net.get', required=['name'])
def host_net_get(agent, msg):
    """
    Get all Networks used by a vim.HostSystem managed object

    Example client message would be:

        {
            "method":     "host.net.get",
            "hostname":   "vc01.example.org",
            "name":       "esxi01.example.org",
        }

    Returns:
        The managed object properties in JSON format

    """
    logger.debug(
        '[%s] Getting Network list available for %s host',
        agent.host,
        msg['name']
    )

    # Find the HostSystem managed object
    # and get the 'network' property
    data = _get_object_properties(
        agent=agent,
        properties=['name', 'network'],
        obj_type=pyVmomi.vim.HostSystem,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    props = data['result'][0]
    host_name, host_networks = props['name'], props['network']

    # Create a list view for the Network managed objects
    view_ref = agent.get_list_view(obj=host_networks)
    result = {}
    result['name'] = host_name
    result['network'] = agent.collect_properties(
        view_ref=view_ref,
        obj_type=pyVmomi.vim.Network,
        path_set=['name']
    )

    view_ref.DestroyView()

    r = {
        'success': 0,
        'msg': 'Successfully discovered objects',
        'result': result,
    }

    return r

@task(name='host.datastore.get', required=['name'])
def host_datastore_get(agent, msg):
    """
    Get all Datastores used by a vim.HostSystem managed object

    Example client message would be:

        {
            "method":   "host.datastore.get",
            "hostname": "vc01.example.org",
            "name":     "esxi01.example.org",
        }

    Returns:
        The discovered objects in JSON format

    """
    return _object_datastore_get(
        agent=agent,
        obj_type=pyVmomi.vim.HostSystem,
        name=msg['name']
    )

@task(name='vm.alarm.get', required=['name'])
def vm_alarm_get(agent, msg):
    """
    Get all alarms for a vim.VirtualMachine managed object

    Example client message would be:

        {
            "method":   "vm.alarm.get",
            "hostname": "vc01.example.org",
            "name":     "vm01.example.org"
        }

    Returns:
        The discovered alarms in JSON format

    """
    result = _object_alarm_get(
        agent=agent,
        obj_type=pyVmomi.vim.VirtualMachine,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    return result

@task(name='vm.perf.metric.get', required=['name', 'counter-id'])
def vm_perf_metric_get(agent, msg):
    """
    Get performance metrics for a vim.VirtualMachine managed object

    Example client message would be:

        {
            "method":   "vm.perf.metric.get",
            "hostname": "vc01.example.org",
            "name":     "vm01.example.org",
            "counter-id"
                12, # CPU Ready time of the Virtual Machine
            ],
            "max-sample": 1,
        }

    For historical performance statistics make sure to pass the
    performance interval key as part of the message, e.g.:

        {
            "method":   "vm.perf.metric.get",
            "hostname": "vc01.example.org",
            "name":     "vm01.example.org",
            "counter-id": [
                12,  # CPU Ready time of the Virtual Machine
                24   # Memory usage as percentage of total configured or available memory
            ],
            "perf-interval": 1 # Historical performance interval '1' (Past day)
        }

    Returns:
        The retrieved performance metrics

    """
    obj = agent.get_object_by_property(
        property_name='name',
        property_value=msg['name'],
        obj_type=pyVmomi.vim.VirtualMachine
    )

    if not obj:
        return {'success': 1, 'msg': 'Cannot find object: {}'.format(msg['name'])}

    try:
        counter_id = int(msg.get('counter-id'))
        max_sample = int(msg.get('max-sample')) if msg.get('max-sample') else 1
        perf_interval_key = int(msg.get('perf-interval')) if msg.get('perf-interval') else None
        instance = msg.get('instance') if msg.get('instance') else ''
    except (TypeError, ValueError):
        logger.warning('Invalid message, cannot retrieve performance metrics')
        return {
            'success': 1,
            'msg': 'Invalid message, cannot retrieve performance metrics'
        }

    return _entity_perf_metric_get(
        agent=agent,
        entity=obj,
        counter_id=counter_id,
        max_sample=max_sample,
        instance=instance,
        perf_interval_key=perf_interval_key
    )

@task(name='vm.perf.metric.info')
def vm_perf_metric_info(agent, msg):
    """
    Get performance counters available for a vim.VirtualMachine object

    Example client message would be:

        {
            "method":     "vm.perf.metric.info",
            "hostname":   "vc01.example.org",
            "name":       "vm01.example.org",
            "counter-id": "<counter-id>"
        }

    Returns:
        Information about the supported performance counters for the object

    """
    obj = agent.get_object_by_property(
        property_name='name',
        property_value=msg['name'],
        obj_type=pyVmomi.vim.VirtualMachine
    )

    if not obj:
        return {'success': 1, 'msg': 'Cannot find object {}'.format(msg['name'])}

    counter_id = int(msg.get('counter-id')) if msg.get('counter-id') else None

    return _entity_perf_metric_info(
        agent=agent,
        entity=obj,
        counter_id=counter_id
    )

@task(name='vm.discover')
def vm_discover(agent, msg):
    """
    Discover all pyVmomi.vim.VirtualMachine managed objects

    Example client message would be:

        {
            "method":   "vm.discover",
            "hostname": "vc01.example.org",
        }

    Example client message which also
    requests additional properties:

        {
            "method":     "vm.discover",
            "hostname":   "vc01.example.org",
            "properties": [
                "name",
                "runtime.powerState"
            ]
        }

    Returns:
        The discovered objects in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    r = _discover_objects(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.VirtualMachine
    )

    return r

@task(name='vm.disk.discover', required=['name'])
def vm_disk_discover(agent, msg):
    """
    Discover all disks used by a vim.VirtualMachine managed object

    Note, that this request requires you to have
    VMware Tools installed in order get information about the
    guest disks.

    Example client message would be:

        {
            "method":   "vm.disk.discover",
            "hostname": "vc01.example.org",
            "name":     "vm01.example.org"
        }

    Example client message requesting
    additional properties to be collected:

        {
            "method":   "vm.disk.discover",
            "hostname": "vc01.example.org",
            "name":     "vm01.example.org",
            "properties": [
                "capacity",
                "diskPath",
                "freeSpace"
            ]
        }

    Returns:
        The discovered objects in JSON format

    """
    logger.debug(
        '[%s] Discovering guest disks for VirtualMachine %s',
        agent.host,
        msg['name']
    )

    # Find the VM and get the guest disks
    data = _get_object_properties(
        agent=agent,
        properties=['name', 'guest.disk'],
        obj_type=pyVmomi.vim.VirtualMachine,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    # Get the VM name and guest disk properties from the result
    props = data['result'][0]
    vm_name, vm_disks = props['name'], props['guest.disk']

    # Properties to be collected for the guest disks
    properties = ['diskPath']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    # Get the requested disk properties
    result = {}
    result['name'] = vm_name
    result['disk'] = [{prop: getattr(disk, prop, '(null)') for prop in properties} for disk in vm_disks]

    r = {
        'success': 0,
        'msg': 'Successfully discovered objects',
        'result': [result],
    }

    return r

@task(name='vm.guest.net.get', required=['name'])
def vm_guest_net_get(agent, msg):
    """
    Discover network adapters for a vim.VirtualMachine  object

    Note, that this request requires you to have
    VMware Tools installed in order get information about the
    guest network adapters.

    Example client message would be:

        {
            "method":   "vm.net.discover",
            "hostname": "vc01.example.org",
            "name":     "vm01.example.org"
        }

    Example client message requesting
    additional properties to be collected:

        {
            "method":   "vm.net.discover",
            "hostname": "vc01.example.org",
            "name":     "vm01.example.org",
            "properties": [
                "network",
                "connected",
                "macAddress",
                "ipConfig"
            ]
        }

    Returns:
        The discovered objects in JSON format

    """
    logger.debug(
        '[%s] Discovering network adapters for VirtualMachine %s',
        agent.host,
        msg['name']
    )

    # Find the VM and get the network adapters
    data = _get_object_properties(
        agent=agent,
        properties=['name', 'guest.net'],
        obj_type=pyVmomi.vim.VirtualMachine,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    # Get the VM name and network adapters
    # properties from the result
    props = data['result'][0]
    vm_name, vm_networks = props['name'], props['guest.net']

    # Properties to be collected for the guest disks
    properties = ['network']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    # Get the requested properties
    result = {}
    result['name'] = vm_name
    result['net'] = [{prop: getattr(net, prop, '(null)') for prop in properties} for net in vm_networks]

    r = {
        'success': 0,
        'msg': 'Successfully retrieved properties',
        'result': result,
    }

    return r

@task(name='vm.net.get', required=['name'])
def vm_net_get(agent, msg):
    """
    Get all Networks used by a vim.VirtualMachine managed object

    Example client message would be:

        {
            "method":     "vm.net.get",
            "hostname":   "vc01.example.org",
            "name":       "vm01.example.org",
        }

    Returns:
        The managed object properties in JSON format

    """
    logger.debug(
        '[%s] Getting Networks available for %s VirtualMachine',
        agent.host,
        msg['name']
    )

    # Find the VirtualMachine managed object and
    # get the 'network' property
    data = _get_object_properties(
        agent=agent,
        properties=['name', 'network'],
        obj_type=pyVmomi.vim.VirtualMachine,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    props = data['result'][0]
    vm_name, vm_networks = props['name'], props['network']

    # Create a list view for the Network managed objects
    view_ref = agent.get_list_view(obj=vm_networks)
    result = {}
    result['name'] = vm_name
    result['network'] = agent.collect_properties(
        view_ref=view_ref,
        obj_type=pyVmomi.vim.Network,
        path_set=['name']
    )

    view_ref.DestroyView()

    r = {
        'success': 0,
        'msg': 'Successfully discovered objects',
        'result': result,
    }

    return r

@task(name='vm.get', required=['name', 'properties'])
def vm_get(agent, msg):
    """
    Get properties for a vim.VirtualMachine managed object

    Example client message would be:

        {
            "method":     "vm.get",
            "hostname":   "vc01.example.org",
            "name":       "vm01.example.org",
            "properties": [
                "name",
                "runtime.powerState"
            ]
        }

    Returns:
        The managed object properties in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    return _get_object_properties(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.VirtualMachine,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

@task(name='vm.host.get', required=['name'])
def vm_host_get(agent, msg):
    """
    Get the vSphere host where a Virtual Machine is running on

    Example client message would be:

        {
            "method":     "vm.host.get",
            "hostname":   "vc01.example.org",
            "name":       "vm01.example.org",
        }

    Returns:
        The managed object properties in JSON format

    """
    logger.debug(
        '[%s] Getting host where %s VirtualMachine is running on',
        agent.host,
        msg['name']
    )

    data = _get_object_properties(
        agent=agent,
        properties=['name', 'runtime.host'],
        obj_type=pyVmomi.vim.VirtualMachine,
        obj_property_name='name',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    props = data['result'][0]
    vm_name, vm_host = props['name'], props['runtime.host']

    result = {
        'name': vm_name,
        'host': vm_host.name,
    }

    r = {
        'success': data['success'],
        'msg': data['msg'],
        'result': [result],
    }

    return r

@task(name='vm.datastore.get', required=['name'])
def vm_datastore_get(agent, msg):
    """
    Get all Datastores used by a vim.VirtualMachine managed object

    Example client message would be:

        {
            "method":   "vm.datastore.get",
            "hostname": "vc01.example.org",
            "name":     "vm01.example.org",
        }

    Returns:
        The discovered objects in JSON format

    """
    return _object_datastore_get(
        agent=agent,
        obj_type=pyVmomi.vim.VirtualMachine,
        name=msg['name']
    )

@task(name='vm.disk.get', required=['name'])
def vm_disk_get(agent, msg):
    """
    Get properties for a disk of a vim.VirtualMachine object

    Note, that this request requires you to have
    VMware Tools installed in order get information about the
    guest disks.

    Example client message would be:

        {
            "method":   "vm.disk.get",
            "hostname": "vc01.example.org",
            "name":     "vm01.example.org"
            "key":      "/var"
        }

    Example client message requesting
    additional properties to be collected:

        {
            "method":   "vm.disk.get",
            "hostname": "vc01.example.org",
            "name":     "vm01.example.org",
            "key":      "/var",
            "properties": [
                "capacity",
                "diskPath",
                "freeSpace"
            ]
        }

    Returns:
        The discovered objects in JSON format

    """
    logger.debug(
        '[%s] Getting guest disk info for %s on VirtualMachine %s',
        agent.host,
        msg['key'],
        msg['name']
    )

    # Discover the VM disks
    data = vm_disk_discover(agent, msg)

    if data['success'] != 0:
        return data

    # If we have no key for the disk,
    # just return the result from discovery
    if 'key' in msg and msg['key']:
        disk_path = msg['key']
    else:
        return data

    props = data['result'][0]
    disks = props['disk']

    for disk in disks:
        if disk['diskPath'] == disk_path:
            break
    else:
        return {
            'success': 1,
            'msg': 'Unable to find guest disk %s' % disk_path
        }

    result = {}
    result['name'] = msg['name']
    result['disk'] = disk

    r = {
        'success': 0,
        'msg': 'Successfully retrieved properties',
        'result': [result],
    }

    return r

@task(name='vm.process.get', required=['name', 'username', 'password'])
def vm_process_get(agent, msg):
    """
    Get processes running on a vim.VirtualMachine managed object

    This method requires you to have VMware Tools installed and
    running in order to get the list of processes running in a
    guest system.

    Example client message would be:

        {
            "method":     "vm.process.get",
            "hostname":   "vc01.example.org",
            "name":       "vm01.example.org",
            "username":   "root",
            "password":   "p4ssw0rd"
        }

    Example client message which requests
    additional properties for the processes:

        {
            "method":     "vm.process.get",
            "hostname":   "vc01.example.org",
            "name":       "vm01.example.org",
            "username":   "root",
            "password":   "p4ssw0rd",
            "properties": [
                "name",
                "owner",
                "pid"
            ]
        }

    Returns:
        The managed object properties in JSON format

    """
    logger.debug(
        '[%s] Getting processes for VirtualMachine %s',
        agent.host,
        msg['name']
    )

    # Get the VirtualMachine managed object
    data = _get_object_properties(
        agent=agent,
        properties=['name', 'guest.toolsRunningStatus'],
        obj_type=pyVmomi.vim.VirtualMachine,
        obj_property_name='name',
        obj_property_value=msg['name'],
        include_mors=True
    )

    if data['success'] != 0:
        return data

    # Get the VM properties
    props = data['result'][0]
    vm_name, vm_tools_is_running, vm_obj = props['name'], props['guest.toolsRunningStatus'], props['obj']

    # Check if we have VMware Tools installed and running first
    # as this request depends on it
    if vm_tools_is_running != 'guestToolsRunning':
        return {
            'success': 1,
            'msg': '%s is not running VMware Tools' % msg['name']
        }

    # Prepare credentials used for
    # authentication in the guest system
    if not msg['username'] or not msg['password']:
        return {'success': 1, 'msg': 'Need username and password for authentication in guest system {}'.format(msg['name'])}

    vm_creds = pyVmomi.vim.vm.guest.NamePasswordAuthentication(
        username=msg['username'],
        password=msg['password']
    )

    try:
        vm_processes = agent.si.content.guestOperationsManager.processManager.ListProcessesInGuest(
            vm=vm_obj,
            auth=vm_creds
        )
    except Exception as e:
        return {
            'success': 1,
            'msg': 'Cannot get guest processes: %s' % e
        }

    # Properties to be collected for the guest processes
    properties = ['cmdLine']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    # Get the requested process properties
    result = [{prop: getattr(process, prop, '(null)') for prop in properties} for process in vm_processes]

    r = {
        'success': 0,
        'msg': 'Successfully retrieved properties',
        'result': result,
    }

    return r

@task(name='vm.cpu.usage.percent', required=['name'])
def vm_cpu_usage_percent(agent, msg):
    """
    Get the CPU usage in percentage for a VirtualMachine

    NOTE: This task will be gone after the transition to performance counters

    Example client message would be:

        {
            "method":     "vm.cpu.usage.percent",
            "hostname":   "vc01.example.org",
            "name":       "vm01.example.org",
        }

    Returns:
        The managed object properties in JSON format

    """
    logger.debug(
        '[%s] Getting CPU usage percentage for VirtualMachine %s',
        agent.host,
        msg['name']
    )

    # Get the VirtualMachine managed object and collect the
    # properties required to calculate the CPU usage percentage.
    # The CPU usage in percentage is directly related to the
    # host the where the Virtual Machine is running on,
    # so we need to collect the 'runtime.host' property as well.
    required_properties = [
        'name',
        'runtime.host',
        'summary.quickStats.overallCpuUsage',
        'config.hardware.numCoresPerSocket',
        'config.hardware.numCPU',
    ]

    data = _get_object_properties(
        agent=agent,
        properties=required_properties,
        obj_type=pyVmomi.vim.VirtualMachine,
        obj_property_name='name',
        obj_property_value=msg['name'],
    )

    if data['success'] != 0:
        return data

    # Get the VM properties
    props = data['result'][0]

    # TODO: A fix for VMware vSphere 4.x versions, where not
    #       always the properties requested are returned by the
    #       vCenter server, which could result in a KeyError
    #       exception. See this issue for more details:
    #
    #           - https://github.com/dnaeon/py-vpoller/issues/33
    #
    #       We should ensure that vPoller Workers do not fail
    #       under such circumstances and return an error message.
    if not all(p in props for p in required_properties):
        return {
            'success': 1,
            'msg': 'Unable to retrieve required properties'
        }

    # Calculate CPU usage in percentage
    # The overall CPU usage returned by vSphere is in MHz, so
    # we first convert it back to Hz and then calculate percentage
    cpu_usage = (
        float(props['summary.quickStats.overallCpuUsage'] * 1048576) /
        (props['runtime.host'].hardware.cpuInfo.hz * props['config.hardware.numCoresPerSocket'] *
         props['config.hardware.numCPU']) *
        100
    )

    result = {
        'name': props['name'],
        'vm.cpu.usage.percent': cpu_usage
    }

    r = {
        'success': 0,
        'msg': 'Successfully retrieved properties',
        'result': [result],
    }

    return r

@task(name='datastore.discover')
def datastore_discover(agent, msg):
    """
    Discover all pyVmomi.vim.Datastore managed objects

    Example client message would be:

        {
            "method":   "datastore.discover",
            "hostname": "vc01.example.org",
        }

    Example client message which also requests
    additional properties:

        {
            "method":     "datastore.discover",
            "hostname":   "vc01.example.org",
            "properties": [
                "name",
                "summary.url"
            ]
        }

    Returns:
        The discovered objects in JSON format

    """
    # Property names to be collected
    properties = ['name']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    r = _discover_objects(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.Datastore
    )

    return r

@task(name='datastore.get', required=['name', 'properties'])
def datastore_get(agent, msg):
    """
    Get properties for a vim.Datastore managed object

    Example client message would be:

        {
            "method":     "datastore.get",
            "hostname":   "vc01.example.org",
            "name":       "ds:///vmfs/volumes/643f118a-a970df28/",
            "properties": [
                "name",
                "summary.accessible",
                "summary.capacity"
            ]
        }

    Returns:
        The managed object properties in JSON format

    """
    # Property names to be collected
    properties = ['name', 'info.url']
    if 'properties' in msg and msg['properties']:
        properties.extend(msg['properties'])

    return _get_object_properties(
        agent=agent,
        properties=properties,
        obj_type=pyVmomi.vim.Datastore,
        obj_property_name='info.url',
        obj_property_value=msg['name']
    )

@task(name='datastore.alarm.get', required=['name'])
def datastore_alarm_get(agent, msg):
    """
    Get all alarms for a vim.Datastore managed object

    Example client message would be:

        {
            "method":   "datastore.alarm.get",
            "hostname": "vc01.example.org",
            "name":     "ds:///vmfs/volumes/643f118a-a970df28/"
        }

    Returns:
        The discovered alarms in JSON format

    """
    result = _object_alarm_get(
        agent=agent,
        obj_type=pyVmomi.vim.Datastore,
        obj_property_name='info.url',
        obj_property_value=msg['name']
    )

    return result

@task(name='datastore.host.get', required=['name'])
def datastore_host_get(agent, msg):
    """
    Get all HostSystem objects attached to a specific Datastore

    Example client message would be:

        {
            "method":     "datastore.host.get",
            "hostname":   "vc01.example.org",
            "name":       "ds:///vmfs/volumes/643f118a-a970df28/",
        }

    """
    logger.info(
        '[%s] Getting HostSystem list using Datastore %s',
        agent.host,
        msg['name']
    )

    # Find the Datastore by it's 'info.url' property
    # and get the HostSystem objects using it
    data = _get_object_properties(
        agent=agent,
        properties=['host'],
        obj_type=pyVmomi.vim.Datastore,
        obj_property_name='info.url',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    # Get properties from the result
    props = data['result'][0]
    obj_host = props['host']

    # obj_host is a list of DatastoreHostMount[] objects,
    # but we need a list of HostSystem ones instead
    obj_host = [h.key for h in obj_host]

    # Get a list view of the hosts from this datastore object
    # and collect their properties
    view_ref = agent.get_list_view(obj=obj_host)
    result = agent.collect_properties(
        view_ref=view_ref,
        obj_type=pyVmomi.vim.HostSystem,
        path_set=['name']
    )

    view_ref.DestroyView()

    r = {
        'success': 0,
        'msg': 'Successfully discovered objects',
        'result': result,
    }

    return r

@task(name='datastore.vm.get', required=['name'])
def datastore_vm_get(agent, msg):
    """
    Get all VirtualMachine objects using a specific Datastore

    Example client message would be:

        {
            "method":     "datastore.vm.get",
            "hostname":   "vc01.example.org",
            "name":       "ds:///vmfs/volumes/643f118a-a970df28/",
        }

    """
    logger.info(
        '[%s] Getting VirtualMachine list using Datastore %s',
        agent.host,
        msg['name']
    )

    # Find the Datastore by it's 'info.url' property and get the
    # VirtualMachine objects using it
    data = _get_object_properties(
        agent=agent,
        properties=['vm'],
        obj_type=pyVmomi.vim.Datastore,
        obj_property_name='info.url',
        obj_property_value=msg['name']
    )

    if data['success'] != 0:
        return data

    # Get properties from the result
    props = data['result'][0]
    obj_vm = props['vm']

    # Get a list view of the VMs from this datastore object
    # and collect their properties
    view_ref = agent.get_list_view(obj=obj_vm)
    result = agent.collect_properties(
        view_ref=view_ref,
        obj_type=pyVmomi.vim.VirtualMachine,
        path_set=['name']
    )

    view_ref.DestroyView()

    r = {
        'success': 0,
        'msg': 'Successfully discovered objects',
        'result': result,
    }

    return r

@task(name='datastore.perf.metric.info', required=['name'])
def datastore_perf_metric_info(agent, msg):
    """
    Get performance counters available for a vim.Datastore object

    Example client message would be:

        {
            "method":     "datastore.perf.metric.info",
            "hostname":   "vc01.example.org",
            "name":       "ds:///vmfs/volumes/643f118a-a970df28/",
            "counter-id": <counter-id>
        }

    Returns:
        Information about the supported performance counters for the object

    """
    obj = agent.get_object_by_property(
        property_name='info.url',
        property_value=msg['name'],
        obj_type=pyVmomi.vim.Datastore
    )

    if not obj:
        return {'success': 1, 'msg': 'Cannot find object %s'.format(msg['name'])}

    counter_id = int(msg.get('counter-id')) if msg.get('counter-id') else None

    return _entity_perf_metric_info(
        agent=agent,
        entity=obj,
        counter_id=counter_id
    )

@task(name='datastore.perf.metric.get', required=['name', 'counter-id'])
def datastore_perf_metric_get(agent, msg):
    """
    Get performance metrics for a vim.Datastore managed object

    The properties passed in the message are the performance
    counter IDs to be retrieved.

    Example client message would be:

        {
            "method":   "datastore.perf.metric.get",
            "hostname": "vc01.example.org",
            "name":     "ds:///vmfs/volumes/643f118a-a970df28/",
            "counter-id": [
                X,
                Y
            ],
            "max_sample": 1,
            "instance": ""
        }

    For historical performance statistics make sure to pass the
    performance interval key as part of the message, e.g.:

        {
            "method":   "datastore.perf.metric.get",
            "hostname": "vc01.example.org",
            "name":     "ds:///vmfs/volumes/643f118a-a970df28/",
            "properties": [
                X,
                Y
            ],
            "key": 1 # Historical performance interval key '1' (Past day)
        }

    Returns:
        The retrieved performance metrics

    """
    obj = agent.get_object_by_property(
        property_name='info.url',
        property_value=msg['name'],
        obj_type=pyVmomi.vim.Datastore
    )

    if not obj:
        return {'success': 1, 'msg': 'Cannot find object: %s' % msg['name']}

    try:
        counter_id = int(msg.get('counter-id'))
        max_sample = int(msg.get('max-sample')) if msg.get('max-sample') else 1
        perf_interval_key = int(msg.get('perf-interval')) if msg.get('perf-interval') else None
        instance = msg.get('instance') if msg.get('instance') else ''
    except (TypeError, ValueError):
        logger.warning('Invalid message, cannot retrieve performance metrics')
        return {
            'success': 1,
            'msg': 'Invalid message, cannot retrieve performance metrics'
        }

    return _entity_perf_metric_get(
        agent=agent,
        entity=obj,
        counter_id=counter_id,
        max_sample=max_sample,
        instance=instance,
        perf_interval_key=perf_interval_key
    )

